#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "gmessage.h"

GMessage *
init_gmessage ()
{
  GMessage    *mesg;
  int         size;

  if (NULL == (mesg = malloc (sizeof(struct _GMessage))) ) {
    fprintf (stderr, "init_gmessage: failed to malloc [1]\n");
    return NULL;
  }

  mesg->cur_len = 0;
  mesg->avail = 0;
  mesg->name = NULL;
  mesg->data = NULL;
  mesg->name_len = NULL;
  mesg->data_len = NULL;
  mesg->serial_len = 0;

  if (0 > (size = gmessage_enlarge (mesg)) ) {
    fprintf (stderr, "init_gmessage: failed to enlarge new gmessage\n");
    deinit_gmessage (mesg);
    return NULL;
  }

  /* printf ("init_gmessage: new gmssage size: %d\n", mesg->cur_len); */

  mesg->is_inited = 1;

  return mesg;
}

void
deinit_gmessage (GMessage *mesg)
{
  int     i;

  if (mesg == NULL)
    return;

  if (mesg->name != NULL && mesg->data != NULL && mesg->cur_len > 0) {
    if (mesg->avail < mesg->cur_len)
      for (i = 0; i < mesg->cur_len; i++)
        if (mesg->name_len[i] > 0)
          gmessage_remove (mesg, (char *)&mesg->name[i], mesg->name_len[i]);

    free (mesg->name);
  }

  mesg->cur_len = 0;
  mesg->avail = 0;

  free (mesg);
}

void
gmessage_put (GMessage *mesg, char *name, int name_len, char *data, int data_len)
{
  int     i = -1, pos = -1;

  if (mesg == NULL || name_len < 1 || name == NULL || data == NULL)
    return;

  if (mesg->is_inited != 1) {
    fprintf (stderr, "gmessage_put: please call init_gmessage(), first\n");
    return;
  }

  if (mesg->avail == 0) {
    if (0 < gmessage_enlarge (mesg)) {
      fprintf (stderr, "gmessage_put: failed to enlarge as required\n");
      return;
    }
  }

  /* printf ("gmessage_put: cur size [1]: mesg->cur_len = %d\n", mesg->cur_len); */

  if (0 > (pos = gmessage_getpos (mesg, name, name_len)) ) {
    for (i = 0, pos = -1; i < mesg->cur_len; i++) {
      if (mesg->name_len[i] == 0) {
        pos = i;
        break;
      }
    }

    if (pos == -1) {
      fprintf (stderr, "gmessage_put: failed to find suitable location\n");
      return;
    }

    if (NULL == (mesg->name[pos] = malloc (sizeof(char) * (name_len+1))) ) {
      fprintf (stderr, "gmessage_put: failed to malloc [1]\n");
      return;
    }
  
    if (NULL == ((char *)mesg->data[pos] = malloc (sizeof(char) * (data_len+1))) ) {
      fprintf (stderr, "gmessage_put: failed to malloc [2]\n");
      free ((char *)mesg->name[pos]);
      return;
    }

    gmessage_strcpy (name, name_len, mesg->name[pos]);
    mesg->name[pos][name_len] = '\0';

    mesg->avail--;
    mesg->name_len[pos] = name_len;
  }
  else {
    if (NULL == ((char *)mesg->data[pos] = realloc ((char *)mesg->data[pos], sizeof(char) * data_len + 1)) ) {
      fprintf (stderr, "gmessage_put: failed to malloc [2]\n");
      free ((char *)mesg->name[pos]);
      return;
    }
  }

  gmessage_strcpy (data, data_len, mesg->data[pos]);
  mesg->data[pos][data_len] = '\0';

  /* printf ("gmessage_put: %s = %s\n", mesg->name[pos], mesg->data[pos]);   */
  /* printf ("gmessage_put: cur size: mesg->cur_len = %d\n", mesg->cur_len); */

  mesg->data_len[pos] = data_len;
}

int
gmessage_get (GMessage *mesg, char *name, int name_len, char *data[])
{
  int     pos;

  if (mesg == NULL || name_len < 1)
    return -1;

  if (mesg->is_inited != 1) {
    fprintf (stderr, "gmessage_get: please call init_gmessage(), first\n");
    return -1;
  }

  if (mesg->avail == mesg->cur_len)
    return -1;

  if (0 > (pos = gmessage_getpos (mesg, name, name_len)) ) {
    fprintf (stderr, "gmessage_get: value could not be found, %s\n", name);
    return -1;
  }

  printf ("gmessage_get: found value '%s' at pos %d\n", name, pos);

  data[0] = mesg->data[pos];
  mesg->get_tmp = mesg->data[pos];

  return mesg->data_len[pos];
}

char *
gmessage_getx (GMessage *mesg, char *name, int name_len, int *data_len)
{
  int     pos;

  if (mesg->is_inited != 1) {
    fprintf (stderr, "gmessage_getx: please call init_gmessage(), first\n");
    return NULL;
  }

  if (0 > (pos = gmessage_getpos (mesg, name, name_len)) ) {
    fprintf (stderr, "gmessage_getx: value could not be found, %s\n", name);
    return NULL;
  }

  data_len = &mesg->data_len[pos];

  return mesg->data[pos];
}

void
gmessage_remove (GMessage *mesg, char *name, int name_len)
{
  int     pos;

  if (mesg == NULL || name == NULL || name_len < 1)
    return;

  if (mesg->is_inited != 1) {
    fprintf (stderr, "gmessage_remove: please call init_gmessage(), first\n");
    return;
  }

  if (0 > (pos = gmessage_getpos (mesg, name, name_len)) )
    return;

  free (mesg->name[pos]);
  free (mesg->data[pos]);

  mesg->name_len[pos] = 0;
  mesg->data_len[pos] = 0;

  mesg->avail += 1;
}

void
gmessage_remove_all (GMessage *mesg)
{
  int     i;

  if (mesg == NULL)
    return;

  if (mesg->is_inited != 1) {
    fprintf (stderr, "gmessage_remove_all: please call init_gmessage(), first\n");
    return;
  }

  if (mesg->avail == mesg->cur_len)
    return;

  for (i = 0; i < mesg->cur_len; i++)
    if (mesg->name_len[i] > 0)
      gmessage_remove (mesg, mesg->name[i], mesg->name_len[i]);
}

int
gmessage_enlarge (GMessage *mesg)
{
  int     amt;

  if (mesg == NULL)
    return -1;

  amt = (int)(mesg->cur_len * 0.5);
  amt += amt < 10 ? 10 : 0;

  printf ("gmessage_enlarge: mesg->cur_len = %d, amt = %d\n", mesg->cur_len, amt);

  if (NULL == (mesg->name = realloc (mesg->name, sizeof(char *) * (mesg->cur_len + amt))) ) {
    fprintf (stderr, "gmessage_enlarge: failed to malloc [1]\n");
    return -1;
  }

  if (NULL == (mesg->data = realloc (mesg->data, sizeof(char *) * (mesg->cur_len + amt))) ) {
    fprintf (stderr, "gmessage_enlarge: failed to malloc [2]\n");
    return -1;
  }

  if (NULL == (mesg->name_len = realloc (mesg->name_len, sizeof(int) * (mesg->cur_len + amt))) ) {
    fprintf (stderr, "gmessage_enlarge: failed to malloc [3]\n");
    return -1;
  }

  if (NULL == (mesg->data_len = realloc (mesg->data_len, sizeof(int) * (mesg->cur_len + amt))) ) {
    fprintf (stderr, "gmessage_enlarge: failed to malloc [4]\n");
    return -1;
  }

  bzero (mesg->name + mesg->cur_len, sizeof(char) * amt);
  bzero (mesg->data + mesg->cur_len, sizeof(char) * amt);

  bzero (mesg->name_len + mesg->cur_len, sizeof(int) * amt);
  bzero (mesg->data_len + mesg->cur_len, sizeof(int) * amt);

  mesg->cur_len += amt;
  mesg->avail += amt;

  /* printf ("gmessage_enlarge: new size: mesg->cur_len = %d\n", mesg->cur_len); */

  return amt;
}

int
gmessage_serialize (GMessage *mesg)
{
  int     i, size, len = 8;
  char    buf[16];

  if (mesg == NULL)
    return -1;

  if (mesg->is_inited != 1) {
    fprintf (stderr, "gmessage_serialize: please call init_gmessage(), first\n");
    return -1;
  }

  if (mesg->serial_len < 8) {
    if (NULL == (mesg->serial = malloc (sizeof(char) * 8)) ) {
      fprintf (stderr, "gmessage_serialize: failed to malloc\n");
      return -1;
    }

    mesg->serial_len += 8;
  }

  memcpy (mesg->serial, "0000000:", 8);

  for (i = 0; i < mesg->cur_len; i++) {
    int     tmp_len, str_len;

    if (mesg->name_len[i] == 0)
      continue;

    tmp_len = mesg->name_len[i] + mesg->data_len[i] + 2;
    str_len = snprintf (buf, 16, "%d", mesg->data_len[i]);

    if (len + tmp_len + str_len > mesg->serial_len) {
      int add_len = (len + tmp_len + str_len) - mesg->serial_len;
      mesg->serial = realloc (mesg->serial, sizeof(char) * (add_len + mesg->serial_len));
      mesg->serial_len += add_len;
    }

    /* printf ("gmessage_serialize [1]: len = %.2d; val = %s\n", len, mesg->serial); */
    gmessage_strcpy (mesg->name[i], mesg->name_len[i], (char *)(mesg->serial + len));
    len += mesg->name_len[i];

    /* printf ("gmessage_serialize [2]: len = %.2d; val = %s\n", len, mesg->serial); */
    gmessage_strcpy (":", 1, (char *)(mesg->serial + len));
    len += 1;

    /* printf ("gmessage_serialize [3]: len = %.2d; val = %s\n", len, mesg->serial); */
    gmessage_strcpy (buf, str_len, (char *)(mesg->serial + len));
    len += str_len;

    /* printf ("gmessage_serialize [4]: len = %.2d; val = %s\n", len, mesg->serial); */
    gmessage_strcpy (":", 1, (char *)(mesg->serial + len));
    len += 1;

    /* printf ("gmessage_serialize [5]: len = %.2d; val = %s\n", len, mesg->serial); */
    gmessage_strcpy (mesg->data[i], mesg->data_len[i], (char *)(mesg->serial + len));
    len += mesg->data_len[i];

    /* printf ("gmessage_serialize [6]: len = %.2d; val = %s\n", len, mesg->serial); */
  }

  if (mesg->serial_len > len)
    mesg->serial = realloc (mesg->serial, len);

  size = snprintf (buf, 16, "%d", mesg->serial_len - 8);
  memcpy (&mesg->serial[7 - size], buf, size);

  /* printf ("gmessage_serialize: %s\n", mesg->serial); */

  return len;
}

GMessage *
gmessage_deserialize (GMessage *mesg, char *data, int data_len)
{
  int     pos1, pos2, len;

  if (mesg->is_inited != 1) {
    fprintf (stderr, "gmessage_deserialize: please call init_gmessage(), first\n");
    return NULL;
  }

  data += 8;
  data_len -= 8;

  for (pos1 = 0; pos1 < data_len; pos1++) {
    if (data[pos1] != ':')
      continue;

    for (pos2 = pos1 + 1; pos2 < data_len; pos2++) {
      if (data[pos2] != ':')
        continue;

      len = gmessage_atoi (data + pos1 + 1, pos2 - pos1);
      gmessage_put (mesg, data, pos1, &data[pos2+1], len);

      data += pos2 + len + 1;
      data_len -= (pos2 + len + 1);
      pos1 = 0;

      break;
    }
  }

  return mesg;
}

int
gmessage_atoi (char *numb, int numb_len)
{
  char numb_str[numb_len+1];

  memcpy (&numb_str, numb, numb_len);
  numb_str[numb_len] = '\0';

  return atoi (numb_str);
}

void
gmessage_strcpy (char *src_str, int src_len, char *dst_str)
{
  int     i;

  for (i = 0; i < src_len; i++)
    *dst_str++ = *src_str++;
}

int
gmessage_strcmp (char *a_str, int a_len, char *b_str, int b_len)
{
  int     i;

  /* printf ("gmessage_strcmp: a_str = %s; a_len = %d\n", a_str, a_len); */
  /* printf ("gmessage_strcmp: b_str = %s; b_len = %d\n", b_str, b_len); */

  if (a_len != b_len)
    return 1;

  /* printf ("gmessage_strcmp: test_str = %s\n", "test_string");           */

  for (i = 0; i < a_len; i++, a_str++, b_str++) {
    /* printf ("gmessage_strcmp: comparing a = %c, b = %c\n", a_str[0], b_str[0]); */
    /* printf ("gmessage_strcmp: a_str = %s; b_str = %s\n", a_str, b_str);         */
    if (*a_str != *b_str)
      return 1;
  }

  return 0;
}

int
gmessage_getpos (GMessage *mesg, char *name, int name_len)
{
  int     i;

  char buf[name_len + 1];
  strncpy (buf, name, name_len);
  buf[name_len] = '\0';

  if (mesg->is_inited != 1) {
    fprintf (stderr, "gmessage_getpos: please call init_gmessage(), first\n");
    return -1;
  }

  /* printf ("gmessage_getpos: searching for %s\n", buf); */

  if (mesg == NULL || name_len < 1)
    return -1;

  for (i = 0; i < mesg->cur_len; i++) {
    /* printf ("gmessage_getpos: testing %d\n", i); */

    if (mesg->name_len[i] == 0)
      continue;

  /* printf ("gmessage_getpos: mesg->name[%d] = %d; mesg->name_len[%d] = %d\n",
                       i, (int)mesg->name[i], i, mesg->name_len[i]);             */

    /* printf ("gmessage_getpos: searching for '%s', found '%s'\n", buf, mesg->name[i]); */

    /* if (0 == gmessage_strcmp ((char *)&mesg->name[i], mesg->name_len[i], name, name_len)) */
    if (0 == gmessage_strcmp (mesg->name[i], mesg->name_len[i], name, name_len))
      return i;
  }

  return -1;
}

