#ifndef __XGMASTER_H
#define __XGMASTER_H

#include <sys/types.h>
#include <sys/socket.h>
#include <pthread.h>

#include "grid.h"
#include "network.h"
#include "xgslave.h"

#define LOCKPOS_H 500
#define LOCKPOS_V 500

typedef struct _GMaster {         /* the groove master       */
  char       *name;               /* server name             */
  int        tcp_port;            /* tcp listen port         */
  int        udp_port;            /* udp listen port         */
  Grid       *grid;               /* display arrangement     */
  GridLoc    *main_screen;        /* master screen           */
  GridLoc    *ctrl_screen;        /* focus screen            */
  GDisplay   *ctrl_display;       /* focus display           */
  GSlave     *ctrl_slave;         /* focus slave             */
  GSlave     *slaves[GRID_SIZE];  /* slave list              */
  GSlave     *slvtmp[GRID_SIZE];  /* un-init slave conns     */
  int        is_locked;           /* 1 = yes, otherwise = no */
  char       *clipboard;          /* clipboard data          */
  int        clipboard_len;       /* clipboard data size     */
  TCPLstn    *tcp_lstn;           /* tcp listener object     */
  UDPConn    *udp_conn;           /* udp listener object     */
  GDisplay   gdisplay;            /* display stuff           */
  pthread_t  tcp_lstn_thread;     /* tcp accept loop         */
  pthread_t  udp_conn_thread;     /* udp 'accept' loop       */
} GMaster;

GMaster * init_gmaster (char *name);
void deinit_gmaster (GMaster *gmaster);

void gmaster_close_conn (GMaster *gmaster, GSlave *gslave);
void do_gmaster (int argc, char *argv[]);
void gmaster_tcp_slave_loop (void *arg);

int gmaster_move (GMaster *gmaster, int dir);
void gmaster_proc_motion (GMaster *gmaster, int pos_x, int pos_y);
void gmaster_proc_click (GMaster *gmaster, int btn, int is_press);
void gmaster_proc_keypress (GMaster *gmaster, int keycode, int is_press);
void gmaster_proc_clipboard (GMaster *gmaster, GClipboardEvent gce);

void gmaster_xevent_selection_send (GMaster *gmaster, XSelectionRequestEvent sre);
void gmaster_xevent_selection_notify (GMaster *gmaster, Window window);

void gmaster_tcp_accept_loop (GMaster *gmaster);
void gmaster_udp_accept_loop (GMaster *gmaster);

void gmaster_lock_mode (GMaster *gmaster);
void gmaster_free_mode (GMaster *gmaster);
void gmaster_setmode_free (GMaster *gmaster);

void gmaster_sig_pipe (int signo);
void gmaster_sig_quit (int signo);   /* tell all of the slaves, goodbye */

#endif /* __XGMASTER_H */

